# -*- coding: utf-8 -*-

#Exercise 3 template for CHE210D

#import python modules
import numpy as np


#import compiled fortran library
import finalPlib
import atomwrite

#whether or not to use the 3d visualization
UseVisual = True
if UseVisual:
    #import custom visualization library
    import atomvis


#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
Cut = 8
#set the random number seed; useful for debugging
np.random.seed(342324)


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.



def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,2) array
    Dir: (N,2) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,2) position array along Dir
"""
    #identify global variables
    global L, Cut, ratio
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [finalPlib.calcenergy(Pos + NormDir * x, L, Cut, ratio) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = finalPlib.calcenergy(Pos + NormDir * dx, L, Cut, ratio)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [finalPlib.calcenergy(Pos + NormDir * 2. * dx, L, Cut, ratio)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [finalPlib.calcenergy(Pos + NormDir * Dists[-1], L, Cut, ratio)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = finalPlib.calcenergy(Pos + NormDir * d3, L, Cut, ratio)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = finalPlib.calcenergy(Pos + NormDir * d3, L, Cut, ratio)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global M, L, Cut
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = finalPlib.calcenergyforces(Pos, L, Cut,ratio, Forces)
    Dir = Forces
    OldPE = 1.e300                                                                    
    #iterative line searches
    k=0
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = finalPlib.calcenergyforces(Pos, L, Cut, ratio, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
        print PE
        if k%10==0:
            Pdb.write(Pos)    
        k=k+1       
    return PE, Pos

def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,2) array of positions
"""
    #make the position array
    Pos = np.zeros((N,2), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./2.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
                Pos[i] = np.array([x,y], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(2) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos

#### YOUR CODE HERE ####

#### YOU WILL NEED TO DEFINE THE FOLLOWING GLOBAL ####
#### VARIABLES FOR THE CODE ABOVE TO WORK         ####

#### finding acceptance ratio in MC steps
"""
drL=[0.07,0.08,0.09,0.1]
ratL=[0.05,0.1,0.2,0.4]
s=0
while s<4:
    N = 1000
    L = (N/0.15)**(1./2.)
    Cut = 10
    ratio=ratL[s]
    dt=0.001
    T=0.05
    Dim=2
    dx=0.001
    EFracTolLS=1.e-8
    EFracTolCG=1.e-4
    dr=drL[s]
    Pos=InitPositions(N, L)
    PEnergy,Pos = ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG)
    print "kk"
    TotE = finalPlib.calcenergy(Pos, L, Cut, ratio)    
    totn=0
    accn=0
    while totn<480000:
        MoveAtom = np.random.randint(len(Pos))
        OldPos = Pos[MoveAtom, :].copy()
        OldTotE=TotE
        OldE = finalPlib.calconeatomenergy(Pos, L, Cut, ratio, MoveAtom)
        Pos[MoveAtom,:] = Pos[MoveAtom,:] + (np.random.rand(1,2)*2-1)*dr
        newE = finalPlib.calconeatomenergy(Pos, L, Cut, ratio, MoveAtom)
        DeltaU=newE-OldE
        P=np.exp(-DeltaU/T)
        a=np.array([1,P],float)
        b=a.min()
        c=np.random.rand(1)
        if b==1: 
            accn=accn + 1
            TotE= TotE + DeltaU
        if b < 1:
            if b >= c: 
                accn=accn+1
                TotE= TotE + DeltaU
            if b < c:
                TotE= OldTotE
                Pos[MoveAtom,:]=OldPos
        totn=totn+1
        if totn%(10*240)==0:
            TotE = finalPlib.calcenergy(Pos, L, Cut, ratio)
        if UseVisual:
            if atomvis.Initialized:
                #update the positions
                atomvis.Update(Pos)
            else:
                #initialize the visualization window
                atomvis.Init(Pos) 
    print dr, float(accn)/float(totn)     
    s=s+1
"""

      
drL=[0.05,0.06,0.07,0.08,0.11]
ratioL=[0.0125,0.025,0.05,0.1,0.2]
q=0
while q<5:   
    N = 750
    L = (N/0.15)**(1./2.)
    Cut = 10
    Cut2= 3
    ratio=ratioL[q]
    dt=0.001
    Dim=2
    T=0.1
    dx=0.001
    EFracTolLS=1.e-8
    EFracTolCG=1.e-10
    Pos=InitPositions(N, L)
    Pdb=atomwrite.pdbfile("anim"+str(ratio)+".pdb")
    PEnergy,Pos = ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG)
    TotE = finalPlib.calcenergy(Pos, L, Cut, ratio)    
    totn=0
    accn=0      
    dr=drL[q]
    f=file("potential_E_ratio"+str(ratio)+"temp_"+str(T)+".txt", "a")
    while totn<480000:
        if totn % N == 0:
            f.write("%10.5f  %10.5f \n" %(float(totn)/750., TotE)) 
        MoveAtom = np.random.randint(len(Pos))
        OldPos = Pos[MoveAtom, :].copy()
        OldTotE=TotE
        OldE = finalPlib.calconeatomenergy(Pos, L, Cut, ratio, MoveAtom)
        Pos[MoveAtom,:] = Pos[MoveAtom,:] + (np.random.rand(1,2)*2-1)*dr
        newE = finalPlib.calconeatomenergy(Pos, L, Cut, ratio, MoveAtom)
        DeltaU=newE-OldE
        P=np.exp(-DeltaU/T)
        a=np.array([1,P],float)
        b=a.min()
        c=np.random.rand(1)
        if b==1: 
            accn=accn + 1
            TotE= TotE + DeltaU
        if b < 1:
            if b >= c: 
                accn=accn+1
                TotE= TotE + DeltaU
            if b < c:
                TotE= OldTotE
                Pos[MoveAtom,:]=OldPos
        totn=totn+1
        if totn % 2400 == 0:
            TotE = finalPlib.calcenergy(Pos, L, Cut, ratio)
            Pdb.write(Pos)
        if UseVisual:
            if atomvis.Initialized:
                    #update the positions
                atomvis.Update(Pos)
            else:
                #initialize the visualization window
                atomvis.Init(Pos)
    print ratio, T, dr, float(accn)/float(totn) 
    f.close()
    s=1 
    totn2=0       
    while s<6:
        f=file("Histogram_s="+str(s)+"ratio="+str(ratio)+"temp="+str(T)+".txt", "w")
        Hist=np.zeros(100,int)
        while totn2<480000*s:
            MoveAtom = np.random.randint(len(Pos))
            OldPos = Pos[MoveAtom, :].copy()
            OldTotE=TotE
            OldE = finalPlib.calconeatomenergy(Pos, L, Cut, ratio, MoveAtom)
            Pos[MoveAtom,:] = Pos[MoveAtom,:] + (np.random.rand(1,2)*2-1)*dr
            newE = finalPlib.calconeatomenergy(Pos, L, Cut, ratio, MoveAtom)
            DeltaU=newE-OldE
            P=np.exp(-DeltaU/T)
            a=np.array([1,P],float)
            b=a.min()
            c=np.random.rand(1)
            if b==1: 
                accn=accn + 1
                TotE= TotE + DeltaU
            if b < 1:
                if b >= c: 
                    accn=accn+1
                    TotE= TotE + DeltaU
                if b < c:
                    TotE= OldTotE
                    Pos[MoveAtom,:]=OldPos
            totn2=totn2+1
            if totn2 % 2400 == 0:
                TotE = finalPlib.calcenergy(Pos, L, Cut, ratio)
                oneAtom=0
                while oneAtom<750:
                    m = finalPlib.calceatomnumber(Pos, L, Cut2, oneAtom)
                    Hist[m]=Hist[m]+1
                    oneAtom=oneAtom+1                        
            if UseVisual:
                if atomvis.Initialized:
                    #update the positions
                    atomvis.Update(Pos)
                else:
                    #initialize the visualization window
                    atomvis.Init(Pos)   
        s=s+1
        Histogram=str(Hist)
        f.write(Histogram)
        f.close()
        pdb.close()  
    q=q+1